pub mod additional_rfactor;
#[allow(warnings)]
mod rfactor_structs;

use std::{fs::File, marker::PhantomData, path::Path};

use additional_rfactor::*;
pub use additional_rfactor::{rF2Vec3, rF2VehicleTelemetry, rF2Wheel};
use anyhow::Result;
use memmap2::Mmap;
pub use rfactor_structs::{ScoringInfoV01, VehicleScoringInfoV01};

const RFACTOR_SHM_FILE: &str = "/dev/shm";

struct ShMReader<T> {
    _file: File,
    shm: Mmap,

    data: PhantomData<T>,
}

impl<T> ShMReader<T> {
    const SIZE: usize = std::mem::size_of::<T>();

    fn new(mm_file_name: &str) -> Result<Self> {
        let file = File::open(Path::new(RFACTOR_SHM_FILE).join(mm_file_name))?;
        let mmap = unsafe { Mmap::map(&file)? };

        Ok(Self {
            _file: file,
            shm: mmap,

            data: PhantomData,
        })
    }
}

impl<'a, T> ShMReader<T>
where
    T: From<&'a [u8]>,
{
    fn read(&'a self) -> T {
        T::from(&self.shm[0..Self::SIZE])
    }
}

pub struct TelemetryReader {
    mm_reader: ShMReader<rF2Telemetry>,
}

impl TelemetryReader {
    pub fn new() -> Result<Self> {
        Ok(Self {
            mm_reader: ShMReader::new(MM_TELEMETRY_FILE_NAME)?,
        })
    }

    pub fn query_telemetry(&self) -> Vec<rF2VehicleTelemetry> {
        self.mm_reader.read().vehicles().to_vec()
    }
}

pub struct ScoringReader {
    mm_reader: ShMReader<rF2Scoring>,
}

impl ScoringReader {
    pub fn new() -> Result<Self> {
        Ok(Self {
            mm_reader: ShMReader::new(MM_SCORING_FILE_NAME)?,
        })
    }

    pub fn vehicle_scoring(&self) -> (ScoringInfoV01, Vec<VehicleScoringInfoV01>) {
        let scoring = self.mm_reader.read();

        (scoring.scoring_info, scoring.vehicles().to_vec())
    }
}
